/*****************************************************************************
 * FILE: compute-kernel.c
 * AUTHOR: William Stafford Noble
 * CREATE DATE: 3/1/99
 * PROJECT: SVM
 * COPYRIGHT: 1999-2001, Columbia University (see ../doc/copyright.html)
 * VERSION: $Revision: 1.32 $
 * DESCRIPTION: Given an m x n matrix of training data (m examples
 *              with n features for each), compute an m x m matrix
 *              of kernel values.
 *****************************************************************************/
#include <stdio.h>
#include <math.h>
#include <assert.h>
#include <string.h>
#include <dlfcn.h>
#include "linear-algebra.h"
#include "matrix.h"
#include "array.h"
#include "class-array.h"
#include "utils.h"
#include "cmdline.h"
#include "rdb-matrix.h"
#include "compute-kernel.h"

#ifndef KERNEL_DEBUG
#define KERNEL_DEBUG 0
#endif

#ifndef RADIAL_DEBUG
#define RADIAL_DEBUG 0
#endif

#ifndef NORMALIZE_DEBUG
#define NORMALIZE_DEBUG 0
#endif

#ifndef DIAG_DEBUG
#define DIAG_DEBUG 0
#endif

/*****************************************************************************
 * Compute a jackknife dot product.
 *
 * Let d_i(x,y) be the dot-product of x and y with the ith element
 * deleted.  Then the jackknife dot-product for two vectors of length
 * N is min{d_1(x,y), d_2(x,y), ..., d_N(x,y), d(x,y)}.
 *****************************************************************************/
double jackknife_dot_product
(ARRAY_T* example1,
 ARRAY_T* example2)
{
  int num_values;
  int i_value;
  double value1;
  double value2;
  double minimum_dot_product;
  double new_dot_product;

  minimum_dot_product = dot_product(example1, example2);
  num_values = get_array_length(example1);
  for (i_value = 0; i_value < num_values; i_value++) {

    /* Store the values. */
    value1 = get_array_item(i_value, example1);
    value2 = get_array_item(i_value, example2);

    /* Replace them with zero. */
    set_array_item(i_value, 0.0, example1);
    set_array_item(i_value, 0.0, example2);

    /* Compute the new dot product. */
    new_dot_product = dot_product(example1, example2);

    /* Have we found a new minimum? */
    if (new_dot_product < minimum_dot_product) {
      minimum_dot_product = new_dot_product;
    }

    /* Put the original values back in. */
    set_array_item(i_value, value1, example1);
    set_array_item(i_value, value2, example2);
  }
  return(minimum_dot_product);
}


/*****************************************************************************
 * Normalize a kernel matrix.
 *
 * Let x be a vector of unnormalized features. Let \tilde{x} be the
 * vector of normalized features. 
 * 
 *         \tilde{x}_i = x_i/||x||
 * 
 * where ||x|| is the norm of x. This means that ||\tilde{x}|| = 1 for
 * all x. What is happening when you normalize in this way is that all
 * feature vectors x are getting their lengths scaled so that they lie
 * on the surface of the unit sphere in the input space.
 * 
 * It turns out that this kind of normalization can be done in general
 * for any kernel. If K(x,y) is any kernel supplied by a user, then
 * you can normalize it by defining
 * 
 *     \tilde{K}(x,y) = K(x,y)/\sqrt{K(x,x)}\sqrt{K(y,y)}
 * 
 * It turns out that \sqrt{K(x,x)} is the norm of x in the feature
 * space.  Hence, for any x, the norm of x in the feature space
 * defined by \tilde{K} is
 * 
 *           \sqrt{\tilde{K}(x,x)}  = 1
 * 
 * When you normalize the kernel this way, it ensures that all points
 * are mapped to the surface of the unit ball in some (possibly infinite
 * dimensional) feature space. 
 *****************************************************************************/
#define SMALL_VALUE  1E-10
static void normalize_kernel_matrix
  (ARRAY_T*  row_self_values,
   ARRAY_T*  col_self_values,
   MATRIX_T* kernel_matrix)
{
  int num_rows;
  int i_row;
  int num_cols;
  int i_col;
  double row_diag;
  double col_diag;
  double this_cell;
  double product;
  static BOOLEAN_T warning = FALSE;

  /* Get the dimensions of the matrix. */
  num_rows = get_num_rows(kernel_matrix);
  assert(num_rows == get_array_length(row_self_values));
  num_cols = get_num_cols(kernel_matrix);
  assert(num_cols == get_array_length(col_self_values));

  for (i_row = 0; i_row < num_rows; i_row++) {
    row_diag = sqrt(get_array_item(i_row, row_self_values));

    for (i_col = 0; i_col < num_cols; i_col++) {
      col_diag = sqrt(get_array_item(i_col, col_self_values));

      /* Get this cell and normalize it. */
      this_cell = get_matrix_cell(i_row, i_col, kernel_matrix);

      /* Check for divide-by-zero. */
      product = row_diag * col_diag;
      if (product == 0.0) {
	if (!warning) {
	  fprintf(stderr, "Warning: Zero-valued diagonal kernel element:");
	  fprintf(stderr, "row %d = %g, col %d = %g.\n", i_row, row_diag,
		  i_col, col_diag);
	  warning = TRUE;
	}
	product = SMALL_VALUE;
      }
	
      if (NORMALIZE_DEBUG) {
	fprintf(stderr, "matrix[%d][%d] = %g / (%g * %g) = %g / %g = %g\n",
		i_row, i_col, this_cell, row_diag, col_diag, 
		this_cell, product, this_cell / product);
      }

      /* Normalize. */
      this_cell /= product;

#ifdef DEBUG
      /* Check for NaN. */
      if (isnan(this_cell)) {
	die("Failure to normalize at [%d,%d] (%g).\n", 
	    i_row, i_col, this_cell);
      }
#endif

      /* Store the normalized result. */
      set_matrix_cell(i_row, i_col, this_cell, kernel_matrix);
    }
  }

  /* Set the self-kernel values to 1.0 */
  init_array(1.0, row_self_values);
  init_array(1.0, col_self_values);
}

/*****************************************************************************
 * Given three parameters, A, B and C, compute (B(X + C))^A.
 *****************************************************************************/
static double polynomialize
  (double    power,
   double    coefficient,
   double    constant,
   double    value)
{
  /* Add the given constant. */
  value += constant;

  /* Multiply by the given coefficient. */
  value *= coefficient;

  /* Raise to the given power. */
  value = pow(value, power);

  return(value);
}

/*****************************************************************************
 * Given three parameters, A, B and C, replace each element X in a
 * given array by the value (B(X + C))^A.
 *****************************************************************************/
static void polynomialize_array
  (double    power,
   double    coefficient,
   double    constant,
   ARRAY_T*  array)
{
  int num_items;
  int i_item;
  double value;

  num_items = get_array_length(array);
  for (i_item = 0; i_item < num_items; i_item++) {

    /* Retrieve the value from the kernel matrix. */
    value = get_array_item(i_item, array);

    /* Polynomialize the kernel value. */
    value = polynomialize(power, coefficient, constant, value);

#ifdef DEBUG
    /* Check for NaN. */
    if (isnan(value)) {
      die("Failure to compute polynomial at [?,%d] (%g).\n",
	  i_item, value);
    }
#endif

    /* Store the new value. */
    set_array_item(i_item, value, array);
  }
}


/*****************************************************************************
 * Given three parameters, A, B and C, replace each element X in a
 * given matrix by the value (B(X + C))^A.  Also perform the same
 * operation on two given arrays.
 *****************************************************************************/
static void polynomialize_matrix
  (double    power,
   double    coefficient,
   double    constant,
   ARRAY_T*  row_self_values,
   ARRAY_T*  col_self_values,
   MATRIX_T* matrix)
{
  int num_rows;
  int i_row;

  /* Polynomialize the matrix. */
  num_rows = get_num_rows(matrix);
  for (i_row = 0; i_row < num_rows; i_row++) {
    polynomialize_array(power, coefficient, constant, 
			get_matrix_row(i_row, matrix));
  }

  /* Polynomialize the self-kernel values. */
  if (row_self_values != NULL) {
    polynomialize_array(power, coefficient, constant, row_self_values);
  }
  if (col_self_values != NULL) {
    polynomialize_array(power, coefficient, constant, col_self_values);
  }
}

/****************************************************************************
 * Define the squared distance as follows:
 *
 *     d^2(x,y) = K(x,x) - 2 K(x,y) + K(y,y)
 *
 * This is the squared Euclidean distance between points in feature
 * space.
 ****************************************************************************/
static double compute_squared_distance
  (double Kxx,
   double Kxy,
   double Kyy)
{
  return(Kxx - (2 * Kxy) + Kyy);
}

/****************************************************************************
 * Set the width of a radial basis kernel to be the median of the
 * distances from each positive example to the nearest negative
 * example.  If no classifications are given, then just use the median
 * of the distance from each example to the nearest other example.
 * 
 * This is only called during training, so we know the kernel matrix
 * is square.
 ****************************************************************************/
static double compute_two_squared_width
  (MATRIX_T*      kernel_matrix,
   double         width_factor,
   CLASS_ARRAY_T* classifications)
{
  int      num_positives;      /* Total number of positive examples. */
  int      num_examples;       /* Number of training examples. */
  int      i_example;          /* Index of this training example. */
  ARRAY_T* nearest_negatives;  /* Distances to nearest negative example. */
  int      i_positive;
  int      i_example1;
  int      i_example2;
  double   squared_distance;   /* Squared distance between two examples. */
  double   return_value;

  num_examples = get_num_rows(kernel_matrix);

  if (classifications == NULL) {
    /* If we ain't got no classifications, treat them all as positives. */
    num_positives = num_examples;
  } else {

    /* Count the number of positive examples. */
    num_positives = 0;
    for (i_example = 0; i_example < num_examples; i_example++) {
      if (get_class(i_example, classifications)) {
	num_positives++;
      }
    }
  }

  /* Allocate the array of distances to nearest negatives. */
  nearest_negatives = allocate_array(num_positives);

  /* Find the nearest negative example for each positive. */
  i_positive = -1;
  for (i_example1 = 0; i_example1 < num_examples; i_example1++) {

    /* Consider only positive examples. */
    if ((classifications == NULL) || 
	(get_class(i_example1, classifications))) {
      i_positive++;

      /* Initialize this position to something very large. */
      set_array_item(i_positive, HUGE_VAL, nearest_negatives);

      for (i_example2 = 0; i_example2 < num_examples; i_example2++) {

	/* Don't look at yourself. */
	if (i_example1 == i_example2) {
	  continue;
	}

	/* Consider only negative examples. */
	if ((classifications == NULL) || 
	    (!get_class(i_example2, classifications))) {

	  /* Compute the distance between these examples. */ 
	  squared_distance
	    = compute_squared_distance(get_matrix_cell(i_example1, 
						       i_example1,
						       kernel_matrix),
				       get_matrix_cell(i_example1,
						       i_example2,
						       kernel_matrix),
				       get_matrix_cell(i_example2, 
						       i_example2,
						       kernel_matrix));

	  /* Store the minimum, ignoring distances of zero. */
	  if ((get_array_item(i_positive, nearest_negatives)
	      > squared_distance) && (squared_distance != 0.0)) {
	    set_array_item(i_positive, squared_distance, nearest_negatives);
	  }
	}
      }
    }
  }

  if (RADIAL_DEBUG) {
    fprintf(stderr, "Distances to negatives: ");
    print_array(nearest_negatives, 5, 2, TRUE, stderr);
  }

  /* Find the median distance. */
  return_value = compute_median(nearest_negatives);
  free_array(nearest_negatives);

  /* Multiply in the given width factor and a factor of 2. */
  return_value *= 2 * width_factor;

  /* If the median minimum distance is zero, fix it. */
  if (return_value == 0.0) {
    return_value = SMALL_VALUE;
  }
  
  /* Return the result. */
  return(return_value);
}

/****************************************************************************
 * Compute a radial basis function kernel, defined by
 *
 *     K(x,y) = exp{-d^2(x,y)/2\sigma^2}
 *
 * for some constant \sigma (the width). 
 ****************************************************************************/
static double radial_kernel
  (double    two_squared_width,
   double    Kxx,
   double    Kxy,
   double    Kyy)
{
  double return_value;
  static BOOLEAN_T seen_underflow = FALSE;

  /* Compute the squared distances between the examples. */
  return_value = compute_squared_distance(Kxx, Kxy, Kyy);

  if (RADIAL_DEBUG) {
    fprintf(stderr, "squared distance=%g\n", return_value);
  }

  /* Divide by twice sigma squared. */
  return_value /= two_squared_width;

  if (RADIAL_DEBUG) {
    fprintf(stderr, "unexponentiated=%g\n", return_value);
  }

  /* Exponentiate the opposite. */
  return_value = exp(-return_value);

  /* Make sure we didn't hit zero. */
  if ((return_value == 0.0) && (!seen_underflow)) {
    fprintf(stderr, "Warning: possible underflow in radial_kernel\n");
    seen_underflow = TRUE;
  }

  if (RADIAL_DEBUG) {
    fprintf(stderr, "return_value=%g\n", return_value);
  }
  return(return_value);
}

/*****************************************************************************
 * Convert each value in a given kernel matrix to a radial basis
 * version.
 *****************************************************************************/
static void radialize_matrix
  (double         constant,
   ARRAY_T*       self_rows,
   ARRAY_T*       self_cols,
   double         two_squared_width,
   MATRIX_T*      kernel_matrix)
{
  int    num_rows;
  int    num_cols;
  int    i_row;
  int    i_col;
  double radial_value;

  /* Get the dimensions of the matrix. */
  num_rows = get_num_rows(kernel_matrix);
  assert(num_rows == get_array_length(self_rows));
  num_cols = get_num_cols(kernel_matrix);
  assert(num_cols == get_array_length(self_cols));

  /* Radialize each row of the matrix. */
  for (i_row = 0; i_row < num_rows; i_row++) {
    for (i_col = 0; i_col < num_cols; i_col++) {

      /* Compute the new value. */
      radial_value
	= radial_kernel(two_squared_width,
			get_array_item(i_row, self_rows),
			get_matrix_cell(i_row, i_col, kernel_matrix),
			get_array_item(i_col, self_cols));

      /* Add the constant back in. */
      radial_value += constant;

#ifdef DEBUG
      /* Check for NaN. */
      if (isnan(radial_value)) {
	die("Failure to radialize at [%d,%d] (%g).\n",
	    i_row, i_col, radial_value);
      }
#endif

      /* Store the new value. */
      set_matrix_cell(i_row, i_col, radial_value, kernel_matrix);
    }
  }
}

/***************************************************************************
 * One good way to set the constants to add to the diagonal is
 *
 *   n+ = number of positive examples
 *   n- = number of negative examples
 *   N  = total number of examples
 *   k  = some constant (given by diagonal_factor)
 *   m  = median value of diagonal
 *
 * Then set
 *
 *   positive_diagonal = (n+/N) * k * m
 *   negative_diagonal = (n-/N) * k * m
 *
 ***************************************************************************/
static void set_diagonal_constants
  (CLASS_ARRAY_T* classes, 
   double         diagonal_factor,
   MATRIX_T*      kernel_matrix,
   double*        positive_diagonal, 
   double*        negative_diagonal)
{
  ARRAY_T* self_kernel_values;
  int num_examples;
  int i_example;
  int num_positive;
  int num_negative;
  double median_diagonal;

  /* If the diagonal factor is zero, don't do nothing. */
  if (diagonal_factor == 0) {
    return;
  }

  /* Find the median self-kernel value. */
  self_kernel_values = extract_diagonal(kernel_matrix);
  median_diagonal = compute_median(self_kernel_values);

  /* Count the number of positives and negatives. */
  num_positive = 0;
  num_negative = 0;
  num_examples = get_num_items(classes);
  for (i_example = 0; i_example < num_examples; i_example++) {
    if (get_class(i_example, classes)) {
      num_positive++;
    } else {
      num_negative++;
    }
  }
  assert(num_positive + num_negative == get_array_length(self_kernel_values));
  free_array(self_kernel_values);

  /* Set the positive and negative diagonal constants. */
  *positive_diagonal = ((double)num_positive / (double)num_examples)
    * diagonal_factor * median_diagonal;
  *negative_diagonal = ((double)num_negative / (double)num_examples)
    * diagonal_factor * median_diagonal;

  if (DIAG_DEBUG) {
    fprintf(stderr, "num_positive=%d num_negative=%d ",
	    num_positive, num_negative);
    fprintf(stderr, "diagonal_factor=%g median_diagonal=%g\n",
	    diagonal_factor, median_diagonal);
    fprintf(stderr, "positive_diagonal=%g negative_diagonal=%g\n",
	    *positive_diagonal, *negative_diagonal);
  }
}

/***************************************************************************
 * Add a constant to the diagonal of the kernel matrix.
 *
 * This can be used to accomplish two things:
 * 
 * (1) If the kernel is not positive definite, then adding a
 *     sufficiently large constant to the diagonal will make it so
 *     (although I don't know how to calculate a priori the proper
 *     value to add).
 *
 * (2) Adding to the diagonal also effectively scales the weights.  A
 *     larger constant makes the weights smaller.  Adding different
 *     constants for the positives and negatives has the same effect
 *     as placing different constraint ceilings.
 *
 ***************************************************************************/
static void add_to_kernel_diagonal
  (CLASS_ARRAY_T* classes, 
   double*        positive_diagonal, 
   double*        negative_diagonal,
   double         diagonal_factor,
   MATRIX_T*      kernel_matrix)
{
  int num_rows;
  int i_row;

  /* Make sure we've got something to do. */
  if (diagonal_factor == 0.0) {
    return;
  }

  /* Use the diagonal factor to compute the positive and negative
     diagonal constants. */
  set_diagonal_constants(classes, diagonal_factor, kernel_matrix,
			 positive_diagonal, negative_diagonal);

  num_rows = get_num_rows(kernel_matrix);
  for (i_row = 0; i_row < num_rows; i_row++) {
    if (get_class(i_row, classes)) {
      incr_matrix_cell(i_row, i_row, *positive_diagonal, kernel_matrix);
    } else {
      incr_matrix_cell(i_row, i_row, *negative_diagonal, kernel_matrix);
    }
  }
}

/****************************************************************************
 * This function effectively subtracts the mean from each feature.
 * 
 * Performing PCA requires centered data; i.e., each FEATURE must have
 * zero mean.  We can't do the centering in the feature space because
 * we don't have access to the feature vectors.  We can, however,
 * express the centered kernel matrix in terms of the non-centered
 * version.
 *
 * Let \phi translate from the input space to the kernel space.  Then
 * we want to find \~K, the centered kernel matrix.  \~K can be
 * expressed in terms of the original matrix K (of dimensionality M)
 * as follows:
 *
 * \~K_{ij} = [\phi(x_i) - 1/M \sum_{m=1}^M \phi(x_m)]
 *             . [\phi(x_j) - 1/M \sum_{n=1}^M \phi(x_n)]
 *
 *          = [\phi(x_i) \phi(x_j)] - [1/M \phi(x_i) \sum_{m=1}^M \phi(x_m)]
 *             - [1/M \phi(x_j) \sum_{n=1}^M \phi(x_n)]
 *             + [1/M^2 \sum_{m=1}^M \sum{n=1}^M \phi(x_m) \phi(x_n)]
 *
 *          = [K_{ij}] - [1/M \sum_{m=1}^M K_{mj}] - [1/M \sum_{n=1}^M K_{in}]
 *             + [1/M^2 \sum_{m=1}^M \sum{n=1}^M K_{mn}]
 *
 * This last, four-term expression is what is computed by this
 * function.  Variables are named according to their place in this
 * expression.
 ****************************************************************************/
static void zero_mean_col_kernel_matrix
  (MATRIX_T* kernel_matrix)
{
  int num_rows;
  int i_row;
  int num_cols;
  int i_col;
  ARRAY_T*  kernel_row_sums;
  ARRAY_T*  kernel_col_sums;
  double    kernel_sum;

  double first_term;     /* K_{ij} */
  double second_term;    /* - 1/M \sum_{m=1}^M k_{mj} */
  double third_term;     /* - 1/M \sum_{n=1}^M k_{in} */
  double fourth_term;    /* 1/M^2 \sum_{m=1}^M \sum{n=1}^M K_{mn} */
  double centered_value; /* The sum of the four terms. */

  /* Get the dimensions of the matrix. */
  num_rows = get_num_rows(kernel_matrix);
  num_cols = get_num_cols(kernel_matrix);

  /* Compute the margins of the matrix. */
  kernel_row_sums = get_matrix_row_sums(kernel_matrix);
  kernel_col_sums = get_matrix_col_sums(kernel_matrix);
  kernel_sum = array_total(kernel_row_sums);

  /* Normalize by the number of rows / columns. */
  scalar_mult(1.0 / (double)(-num_rows), kernel_row_sums);
  scalar_mult(1.0 / (double)(-num_rows), kernel_col_sums);
  kernel_sum /= (double)(num_rows * num_rows);

  /* Update each cell of the kernel matrix. */
  for (i_row = 0; i_row < num_rows; i_row++) {
    for (i_col = 0; i_col < num_cols; i_col++) {

      /* Compute and store all four terms. */
      first_term = get_matrix_cell(i_row, i_col, kernel_matrix);
      second_term = get_array_item(i_row, kernel_row_sums);
      third_term = get_array_item(i_col, kernel_col_sums);
      fourth_term = kernel_sum;
      centered_value = first_term + second_term + third_term + fourth_term;

      /* Update the matrix cell. */
      set_matrix_cell(i_row, i_col, centered_value, kernel_matrix);
    }
  }
    
  free_array(kernel_row_sums);
  free_array(kernel_col_sums);
}

/*****************************************************************************
 * Perform the following operations on the given kernel matrix:
 *
 *  - add constant to diagonal
 *  - normalize
 *  - polynomialize
 *  - radialize
 *  - add asymmetrically to diagonal
 *  - set features to have zero mean
 *
 *****************************************************************************/
void transform_kernel_matrix
  (BOOLEAN_T      square, // Is this a square kernel matrix for one data set.
   KERNEL_T*      kernel,
   CLASS_ARRAY_T* classes,
   ARRAY_T*       self_train,
   ARRAY_T*       self_test,
   MATRIX_T*      kernel_matrix)
{
  if (verbosity >= NORMAL_VERBOSE) {
    fprintf(stderr, "Transforming the kernel matrix.\n");
  }

  // Add to the diagonal, if requested.
  if ((kernel->add_diag != 0.0) && square) {
    if (verbosity > NORMAL_VERBOSE) {
      fprintf(stderr, "Adding %g to kernel diagonal.\n", kernel->add_diag);
    }

    if (square) {
      add_to_diagonal(kernel->add_diag, kernel_matrix);
    }

    scalar_add(kernel->add_diag, self_train);
    scalar_add(kernel->add_diag, self_test);
  }

  /* Normalize the length of each vector. */
  if (kernel->normalize) {
    if (verbosity > NORMAL_VERBOSE) {
      fprintf(stderr, "Normalizing the kernel matrix.\n");
    }
    normalize_kernel_matrix(self_test, self_train, kernel_matrix);
  }

  /* Compute polynomial version of the matrix. */
  if (verbosity > NORMAL_VERBOSE) {
    fprintf(stderr, "Polynomializing the kernel matrix.\n");
  }
  polynomialize_matrix(kernel->power, 
		       kernel->coefficient, 
		       kernel->constant, 
		       self_train, 
		       self_test,
		       kernel_matrix);

  /* Convert to a radial basis kernel. */
  if (kernel->radial) {
    if (verbosity > NORMAL_VERBOSE) {
      fprintf(stderr, "Radializing the kernel matrix.\n");
    }

    /* Compute the kernel width, if it wasn't given. */
    if (kernel->two_squared_width == 0.0) {
      kernel->two_squared_width 
	= compute_two_squared_width(kernel->kernel_matrix, 
				    kernel->width_factor,
				    classes);
    }
    radialize_matrix(kernel->constant, 
		     self_test, 
		     self_train, 
		     kernel->two_squared_width,
		     kernel_matrix);
  }

  /* Add constants to the diagonal. */
  if ((kernel->diagonal_factor != 0.0) && square) {
    if (verbosity > NORMAL_VERBOSE) {
      fprintf(stderr, "Adding to kernel diagonal.\n");
    }
    add_to_kernel_diagonal(classes, 
			   &(kernel->positive_diagonal),
			   &(kernel->negative_diagonal),
			   kernel->diagonal_factor,
			   kernel_matrix);
  }

  /* Subtract the mean from each feature. */
  if (kernel->zero_mean_col) {
    if (verbosity > NORMAL_VERBOSE) {
      fprintf(stderr, "Subtracting the mean from each feature.\n");
    }
    zero_mean_col_kernel_matrix(kernel_matrix);
  }
}

/*****************************************************************************
 * Given two input matrices, compute a matrix of pairwise dot products.
 *****************************************************************************/
void compute_base_kernel_matrix
(MATRIX_T* matrix1, // test if in classify 
 MATRIX_T* matrix2, // train
   MATRIX_T* kernel_matrix)
{
  ARRAY_T* example1;
  ARRAY_T* example2;
  int num_rows;
  int num_cols;
  int i_row;
  int i_col;
  double kernel;

  if (verbosity >= NORMAL_VERBOSE) {
    fprintf(stderr, "Computing base kernel matrix.\n");
  }

  /* Compute the size of the output matrix. */
  num_rows = get_num_rows(matrix1);
  assert(get_num_rows(kernel_matrix) == num_rows);
  num_cols = get_num_rows(matrix2);
  assert(get_num_cols(kernel_matrix) == num_cols);
  assert(get_num_cols(matrix1) == get_num_cols(matrix2));

  /* Compute the kernel for every pair of examples. */
  for (i_row = 0; i_row < num_rows; i_row++) {
    example1 = get_matrix_row(i_row, matrix1);

    if (verbosity >= DUMP_VERBOSE) {
      fprintf(stderr, "Computing kernels for example %d.\n", i_row);
    }

    for (i_col = 0; i_col < num_cols; i_col++) {
      example2 = get_matrix_row(i_col, matrix2);

      /* Compute the dot product. */
      kernel = dot_product(example1, example2);

#ifdef DEBUG
      /* Check for NaN. */
      if (isnan(kernel)) {
	die("Failure to compute base kernel at [%d,%d] (%g).\n",
	    i_row, i_col, kernel);
      }
#endif

      /* Store the resulting value in the matrix. */
      set_matrix_cell(i_row, i_col, kernel, kernel_matrix);

      if (KERNEL_DEBUG) {
	fprintf(stderr, "K(%d,%d)=%g\n", i_row, i_col, kernel);
      }
    }
  }
}


/*
 * Local Variables:
 * mode: c
 * c-basic-offset: 2
 * End:
 */


